/*
 * SPDX-FileCopyrightText: 2024-2026 Sébastien Helleu <flashcode@flashtux.org>
 *
 * SPDX-License-Identifier: GPL-3.0-or-later
 *
 * This file is part of WeeChat, the extensible chat client.
 *
 * WeeChat is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * WeeChat is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with WeeChat.  If not, see <https://www.gnu.org/licenses/>.
 */

/* Test hook process functions */

#include "CppUTest/TestHarness.h"

#include "tests.h"

extern "C"
{
#include "src/core/weechat.h"
}

TEST_GROUP(HookProcess)
{
};

/*
 * Test functions:
 *   hook_process_get_description
 */

TEST(HookProcess, GetDescription)
{
    /* TODO: write tests */
}

/*
 * Test functions:
 *   hook_process_hashtable
 */

TEST(HookProcess, Hashtable)
{
    /* TODO: write tests */
}

/*
 * Test functions:
 *   hook_process
 */

TEST(HookProcess, Process)
{
    /* TODO: write tests */
}

/*
 * Test functions:
 *   hook_process_child
 */

TEST(HookProcess, Child)
{
    /* TODO: write tests */
}

/*
 * Test functions:
 *   hook_process_send_buffers
 */

TEST(HookProcess, SendBuffers)
{
    /* TODO: write tests */
}

/*
 * Test functions:
 *   hook_process_add_to_buffer
 */

TEST(HookProcess, AddToBuffer)
{
    /* TODO: write tests */
}

/*
 * Test functions:
 *   hook_process_child_read
 */

TEST(HookProcess, ChildRead)
{
    /* TODO: write tests */
}

/*
 * Test functions:
 *   hook_process_child_read_stdout_cb
 */

TEST(HookProcess, ChildReadStdoutCb)
{
    /* TODO: write tests */
}

/*
 * Test functions:
 *   hook_process_child_read_stderr_cb
 */

TEST(HookProcess, ChildReadStderrCb)
{
    /* TODO: write tests */
}

/*
 * Test functions:
 *   hook_process_child_read_until_eof
 */

TEST(HookProcess, ChildReadUntilEof)
{
    /* TODO: write tests */
}

/*
 * Test functions:
 *   hook_process_timer_cb
 */

TEST(HookProcess, TimerCb)
{
    /* TODO: write tests */
}

/*
 * Test functions:
 *   hook_process_run
 */

TEST(HookProcess, Run)
{
    /* TODO: write tests */
}

/*
 * Test functions:
 *   hook_process_exec
 */

TEST(HookProcess, Exec)
{
    /* TODO: write tests */
}

/*
 * Test functions:
 *   hook_process_free_data
 */

TEST(HookProcess, FreeData)
{
    /* TODO: write tests */
}

/*
 * Test functions:
 *   hook_process_add_to_infolist
 */

TEST(HookProcess, AddToInfolist)
{
    /* TODO: write tests */
}

/*
 * Test functions:
 *   hook_process_print_log
 */

TEST(HookProcess, PrintLog)
{
    /* TODO: write tests */
}
